@@ -4,16 +4,16 @@ Callable objects that generate numbers according to different distributions.
 
 import random
 import operator
+import hashlib
+import struct
+import fractions
 
 from math import e,pi
 
 import param
 
 
-from param.version import Version
-__version__ = Version(release=(1,2,1), fpath=__file__, 
-                      commit="$Format:%h$", reponame="param")
-
+from param import __version__  # noqa: API import
 
 class TimeAware(param.Parameterized):
     """
@@ -120,6 +120,23 @@ class NumberGenerator(param.Parameterized):
     def __abs__ (self): return UnaryOperator(self,operator.abs)
 
 
+operator_symbols = {
+    operator.add:'+',
+    operator.sub:'-',
+    operator.mul:'*',
+    operator.mod:'%',
+    operator.pow:'**',
+    operator.truediv:'/',
+    operator.floordiv:'//',
+    operator.neg:'-',
+    operator.pos:'+',
+    operator.abs:'abs',
+}
+
+def pprint(x, *args, **kwargs):
+    "Pretty-print the provided item, translating operators to their symbols"
+    return x.pprint(*args, **kwargs) if hasattr(x,'pprint') else operator_symbols.get(x, repr(x))
+
 
 class BinaryOperator(NumberGenerator):
     """Applies any binary operator to NumberGenerators or numbers to yield a NumberGenerator."""
@@ -148,6 +165,10 @@ class BinaryOperator(NumberGenerator):
         return self.operator(self.lhs() if callable(self.lhs) else self.lhs,
                              self.rhs() if callable(self.rhs) else self.rhs, **self.args)
 
+    def pprint(self, *args, **kwargs):
+        return (pprint(self.lhs,      *args, **kwargs) +
+                pprint(self.operator, *args, **kwargs) +
+                pprint(self.rhs,      *args, **kwargs))
 
 
 class UnaryOperator(NumberGenerator):
@@ -171,41 +192,111 @@ class UnaryOperator(NumberGenerator):
     def __call__(self):
         return self.operator(self.operand(),**self.args)
 
+    def pprint(self, *args, **kwargs):
+        return (pprint(self.operator, *args, **kwargs) + '(' +
+                pprint(self.operand,  *args, **kwargs) + ')')
 
-class RandomDistribution(NumberGenerator, TimeAware):
+
+class Hash(object):
     """
-    Python's random module provides the Random class, which can be
-    instantiated to give an object that can be asked to generate
-    numbers from any of several different random distributions
-    (e.g. uniform, Gaussian).
+    A platform- and architecture-independent hash function (unlike
+    Python's inbuilt hash function) for use with an ordered collection
+    of rationals or integers.
 
-    To make it easier to use these, Numbergen provides here a
-    hierarchy of classes, each tied to a particular random
-    distribution. RandomDistributions support setting parameters on
-    creation rather than passing them each call, and allow pickling to
-    work properly.  Code that uses these classes will be independent
-    of how many parameters are used by the underlying distribution,
-    and can simply treat them as a generic source of random numbers.
+    The supplied name sets the initial hash state.  The output from
+    each call is a 32-bit integer to ensure the value is a regular
+    Python integer (and not a Python long) on both 32-bit and 64-bit
+    platforms. This can be important to seed Numpy's random number
+    generator safely (a bad Numpy bug!).
 
-    The underlying random.Random() instance and all its methods can be
-    accessed from the 'random_generator' attribute.
+    The number of inputs (integer or rational numbers) to be supplied
+    for __call__ must be specified in the constructor and must stay
+    constant across calls.
+    """
+    def __init__(self, name, input_count):
+        self.name = name
+        self.input_count = input_count
+        self._digest = hashlib.md5()
+        self._digest.update(name.encode())
+        self._hash_struct = struct.Struct( "!" +" ".join(["I"] * (input_count * 2)))
 
-    RandomDistributions are TimeAware, and thus can be locked to
-    a global time if desired.  By default, time_dependent=False, and
-    so a new random value will be generated each time these objects
-    are called.  If you have a global time function, you can set
-    time_dependent=True, so that the random values will instead be
-    constant at any given time, changing only when the time changes.
-    Using time_dependent values can help you obtain fully reproducible
-    streams of random numbers, even if you e.g. move time forwards and
-    backwards for testing.
 
-    If declared time_dependent, a hash is generated for seeding the
-    random state on each call, using a triple consisting of the object
-    name, the time returned by time_fn and the global value of
-    param.random_seed. As a consequence, for a given name and fixed
-    value of param.random_seed, the random values generated will be a
-    fixed function of time.
+    def _rational(self, val):
+        """Convert the given value to a rational, if necessary."""
+
+        I32 = 4294967296 # Maximum 32 bit unsigned int (i.e. 'I') value
+        if isinstance(val, int):
+            numer, denom = val, 1
+        elif isinstance(val, fractions.Fraction):
+            numer, denom = val.numerator, val.denominator
+        elif hasattr(val, 'numer'):
+            (numer, denom) = (int(val.numer()), int(val.denom()))
+        else:
+            param.main.param.log(param.WARNING, "Casting type '%s' to Fraction.fraction"
+                               % type(val).__name__)
+            frac = fractions.Fraction(str(val))
+            numer, denom = frac.numerator, frac.denominator
+        return numer % I32, denom % I32
+
+
+    def __getstate__(self):
+        """
+        Avoid Hashlib.md5 TypeError in deepcopy (hashlib issue)
+        """
+        d = self.__dict__.copy()
+        d.pop('_digest')
+        d.pop('_hash_struct')
+        return d
+
+
+    def __setstate__(self, d):
+        self._digest = hashlib.md5()
+        name, input_count = d['name'], d['input_count']
+        self._digest.update(name.encode())
+        self._hash_struct = struct.Struct( "!" +" ".join(["I"] * (input_count * 2)))
+        self.__dict__.update(d)
+
+
+    def __call__(self, *vals):
+        """
+        Given integer or rational inputs, generate a cross-platform,
+        architecture-independent 32-bit integer hash.
+        """
+        # Convert inputs to (numer, denom) pairs with integers
+        # becoming (int, 1) pairs to match gmpy.mpqs for int values.
+        pairs = [self._rational(val) for val in vals]
+        # Unpack pairs and fill struct with ints to update md5 hash
+        ints = [el for pair in pairs for el in pair]
+        digest = self._digest.copy()
+        digest.update(self._hash_struct.pack(*ints))
+        # Convert from hex string to 32 bit int
+        return int(digest.hexdigest()[:7], 16)
+
+
+
+class TimeAwareRandomState(TimeAware):
+    """
+    Generic base class to enable time-dependent random
+    streams. Although this class is the basis of all random numbergen
+    classes, it is designed to be useful whenever time-dependent
+    randomness is needed using param's notion of time. For instance,
+    this class is used by the imagen package to define time-dependent,
+    random distributions over 2D arrays.
+
+    For generality, this class may use either the Random class from
+    Python's random module or numpy.random.RandomState. Either of
+    these random state objects may be used to generate numbers from
+    any of several different random distributions (e.g. uniform,
+    Gaussian). The latter offers the ability to generate
+    multi-dimensional random arrays and more random distributions but
+    requires numpy as a dependency.
+
+    If declared time_dependent, the random state is fully determined
+    by a hash value per call. The hash is initialized once with the
+    object name and then per call using a tuple consisting of the time
+    (via time_fn) and the global param.random_seed.  As a consequence,
+    for a given name and fixed value of param.random_seed, the random
+    values generated will be a fixed function of time.
 
     If the object name has not been set and time_dependent is True, a
     message is generated warning that the default object name is
@@ -215,8 +306,121 @@ class RandomDistribution(NumberGenerator, TimeAware):
     explicitly when you construct the RandomDistribution object.
     """
 
+    random_generator = param.Parameter(
+        default=random.Random((500,500)), doc=
+        """
+        Random state used by the object. This may may be an instance
+        of random.Random from the Python standard library or an
+        instance of numpy.random.RandomState.
+
+        This random state may be exclusively owned by the object or
+        may be shared by all instance of the same class. It is always
+        possible to give an object its own unique random state by
+        setting this parameter with a new random state instance.
+        """)
+
     __abstract = True
 
+    def _initialize_random_state(self, seed=None, shared=True, name=None):
+        """
+        Initialization method to be called in the constructor of
+        subclasses to initialize the random state correctly.
+
+        If seed is None, there is no control over the random stream
+        (no reproducibility of the stream).
+
+        If shared is True (and not time-dependent), the random state
+        is shared across all objects of the given class. This can be
+        overridden per object by creating new random state to assign
+        to the random_generator parameter.
+        """
+        if seed is None: # Equivalent to an uncontrolled seed.
+            seed = random.Random().randint(0, 1000000)
+            suffix = ''
+        else:
+            suffix = str(seed)
+
+        # If time_dependent, independent state required: otherwise
+        # time-dependent seeding (via hash) will affect shared
+        # state. Note that if all objects have time_dependent=True
+        # shared random state is safe and more memory efficient.
+        if self.time_dependent or not shared:
+            self.random_generator = type(self.random_generator)(seed)
+
+        # Seed appropriately (if not shared)
+        if not shared:
+            self.random_generator.seed(seed)
+
+        if name is None:
+            self._verify_constrained_hash()
+
+        hash_name = name if name else self.name
+        if not shared:  hash_name += suffix
+        self._hashfn = Hash(hash_name, input_count=2)
+
+        if self.time_dependent:
+            self._hash_and_seed()
+
+
+    def _verify_constrained_hash(self):
+        """
+        Warn if the object name is not explicitly set.
+        """
+        changed_params = self.param.values(onlychanged=True)
+        if self.time_dependent and ('name' not in changed_params):
+            self.param.log(param.WARNING, "Default object name used to set the seed: "
+                           "random values conditional on object instantiation order.")
+
+    def _hash_and_seed(self):
+        """
+        To be called between blocks of random number generation. A
+        'block' can be an unbounded sequence of random numbers so long
+        as the time value (as returned by time_fn) is guaranteed not
+        to change within the block. If this condition holds, each
+        block of random numbers is time-dependent.
+
+        Note: param.random_seed is assumed to be integer or rational.
+        """
+        hashval = self._hashfn(self.time_fn(), param.random_seed)
+        self.random_generator.seed(hashval)
+
+
+
+class RandomDistribution(NumberGenerator, TimeAwareRandomState):
+    """
+    The base class for all Numbergenerators using random state.
+
+    Numbergen provides a hierarchy of classes to make it easier to use
+    the random distributions made available in Python's random module,
+    where each class is tied to a particular random distribution.
+
+    RandomDistributions support setting parameters on creation rather
+    than passing them each call, and allow pickling to work properly.
+    Code that uses these classes will be independent of how many
+    parameters are used by the underlying distribution, and can simply
+    treat them as a generic source of random numbers.
+
+    RandomDistributions are examples of TimeAwareRandomState, and thus
+    can be locked to a global time if desired.  By default,
+    time_dependent=False, and so a new random value will be generated
+    each time these objects are called.  If you have a global time
+    function, you can set time_dependent=True, so that the random
+    values will instead be constant at any given time, changing only
+    when the time changes.  Using time_dependent values can help you
+    obtain fully reproducible streams of random numbers, even if you
+    e.g. move time forwards and backwards for testing.
+
+    Note: Each RandomDistribution object has independent random state.
+    """
+
+    seed = param.Integer(default=None, allow_None=True, doc="""
+       Sets the seed of the random number generator and can be used to
+       randomize time dependent streams.
+
+       If seed is None, there is no control over the random stream
+       (i.e. no reproducibility of the stream).""")
+
+    __abstract = True
 
     def __init__(self,**params):
         """
@@ -224,46 +428,17 @@ class RandomDistribution(NumberGenerator, TimeAware):
         positional and keyword arguments.
 
         If seed=X is specified, sets the Random() instance's seed.
-        Otherwise, calls the instance's jumpahead() method to get a
-        state very likely to be different from any just used.
+        Otherwise, calls creates an unseeded Random instance which is
+        likely to result in a state very different from any just used.
         """
-        self.random_generator = random.Random()
-
-        seed = params.pop('seed', None)
         super(RandomDistribution,self).__init__(**params)
-
-        if seed is not None:
-            self.random_generator.seed(seed)
-        else:
-            self.random_generator.jumpahead(10)
-
-        self._verify_constrained_hash()
-        if self.time_dependent:
-            self._hash_and_seed()
-
-
-    def _verify_constrained_hash(self):
-        changed_params = dict(self.get_param_values(onlychanged=True))
-        if self.time_dependent and ('name' not in changed_params):
-            self.warning("Default object name used to set the seed: "
-                         "random values conditional on object instantiation order.")
-
-    def _hash_and_seed(self):
-        time = self.time_fn()
-        if hasattr(time, 'numer'):
-            time = (int(time.numer()), int(time.denom()))
-        elif not isinstance(time, int):
-            self.warning("Cannot generate known hash format for time type '%s'" % type(time).__name__)
-        hashval = hash((self.name, time, param.random_seed))
-        self.random_generator.seed(hashval)
-
+        self._initialize_random_state(seed=self.seed, shared=False)
 
     def __call__(self):
         if self.time_dependent:
             self._hash_and_seed()
 
 
-
 class UniformRandom(RandomDistribution):
     """
     Specified with lbound and ubound; when called, return a random
@@ -451,7 +626,7 @@ class SquareWave(NumberGenerator, TimeDependent):
          Duration of the 'on' state during which a value of 1.0 is
          returned.""")
 
-    off_duration = param.Number(default=None, allow_None=True, 
+    off_duration = param.Number(default=None, allow_None=True,
                                 bounds=(0.0,None), doc="""
         Duration of the 'off' value state during which a value of 0.0
         is returned. By default, this duration matches the value of
@@ -503,8 +678,8 @@ class ExponentialDecay(NumberGenerator, TimeDependent):
     def __call__(self):
         Vi = self.starting_value
         Vm = self.ending_value
-        return Vm + (Vi - Vm) * self.base**(-1.0*float(self.time_fn())/
-                                                 float(self.time_constant))
+        exp = -1.0*float(self.time_fn())/float(self.time_constant)
+        return Vm + (Vi - Vm) * self.base**exp
 
 
 
@@ -515,11 +690,11 @@ class TimeSampledFn(NumberGenerator, TimeDependent):
     held constant within each interval.
     """
 
-    period = param.Number(default=1.0, bounds=(0.0,None), 
+    period = param.Number(default=1.0, bounds=(0.0,None),
         inclusive_bounds=(False,True), softbounds=(0.0,5.0), doc="""
         The periodicity with which the values of fn are sampled.""")
 
-    offset = param.Number(default=0.0, bounds=(0.0,None), 
+    offset = param.Number(default=0.0, bounds=(0.0,None),
                           softbounds=(0.0,5.0), doc="""
         The offset from time 0.0 at which the first sample will be drawn.
         Must be less than the value of period.""")
@@ -572,8 +747,8 @@ class BoundedNumber(NumberGenerator):
     def __call__(self):
         val = self.generator()
         min_, max_ = self.bounds
-        if   min_ != None and val < min_: return min_
-        elif max_ != None and val > max_: return max_
+        if   min_ is not None and val < min_: return min_
+        elif max_ is not None and val > max_: return max_
         else: return val
 
 
